func_var_multi<-function(x)
{
  # fx=x[1]^4+x[1]^3-x[1]+x[2]^4-x[2]^2 +x[2]+x[3]^2-x[3]+x[1]*x[2]*x[3]
  # fx=0.5*x[1]^2+x[2]^2+x[1]+0.5*x[2]+3
  #fx=garch11_likelihood(x,data)
  fx=x[1]^2+x[2]^2+x[3]^2
  return(fx)
}

# compute the gradient numerically

grad_vec<-function(x)
{ delx=0.0001
grad_vec=c()
for(i in 1:length(x))
{ x_old=x
x_new=x
x_old=x
x_old[i]=x[i]-delx
x_new[i]=x[i]+delx
grad_vec[i]=(func_var_multi(x_new)-func_var_multi(x_old))/(2*delx)
}
return(grad_vec)  
}


##################################################
#
#
#

golden_line_search<-function(a,b,x,sig,epsilon,max_iter)
{
  tau = 0.381967   # golden section
  alpha1 = a*(1-tau) + b*tau
  alpha2 = a*tau + b*(1-tau)
  falpha1 = func_var_multi(x+alpha1*sig)
  falpha2 = func_var_multi(x+alpha2*sig)
  for (k in 1:max_iter)
  {
    if(falpha1>falpha2)
    {
      a = alpha1
      alpha1 = alpha2
      falpha1 = falpha2
      alpha2 = tau*a + (1-tau)*b
      falpha2 = func_var_multi(x+alpha2*sig) 
      
    }
    else
    {
      b = alpha2
      alpha2 = alpha1
      falpha2 = falpha1
      alpha1 = tau*b + (1-tau)*a
      falpha1 = func_var_multi(x+alpha1*sig)
    }
    if (abs(func_var_multi(x+alpha1*sig)-func_var_multi(x+alpha2*sig))
        < epsilon)
    {break}
  }
  golden=c(alpha1 , falpha1)
  return(golden)
}



#########################################
#
#  Steepest descent
#

steepest_descent_standard<-function(x,epsilon,alpha,max_iter)
{
  f_old=func_var_multi(x)
  n_iter=0   # counter of iteration number
  for(k in 0:max_iter)
  { 
    n_iter=n_iter+1
    grad<-grad_vec(x)
    s_dir<- -grad   # compute the gradient and set search direction -Df
    
    # compute the step length using the golden search rule
    
    
    
    x=x+alpha*s_dir
    f_new=func_var_multi(x)
    
    # if f_old and f_new are close  or
    # the gradient is close to 0 we terminate the algorithm
    
    if(abs(f_new-f_old)<epsilon | norm(grad,"2")<epsilon)
      # if(norm(grad,"2")<epsilon)
    {break}
    norm_grad=norm(grad,"2")
   f_old<-f_new
    x=x+alpha*s_dir
  }
  grad<-grad_vec(x)
  norm_grad=norm(grad,"2")
  xm=x
  gradm=grad_vec(xm)
  norm_gradm=norm(gradm,"2")
  results=list(xm,norm_gradm,n_iter)
  return(results)
}


steepest_descent_golden_search<-function(x,epsilon,max_iter)
{
  f_old=func_var_multi(x)
  n_iter=0
  for(k in 0:max_iter)
  {
    n_iter=n_iter+1
    grad<-grad_vec(x)
    s_dir<- -grad   # compute the gradient and set search direction -Df
    
    # compute the step length using the golden search rule
    
    golden=golden_line_search(0,100,x,s_dir,epsilon,max_iter)
    alpha<-golden[1]
    f_new<-golden[2]
    
    # if f_old and f_new are close  or
    # the gradient is close to 0 we terminate the algorithm
    
    if(abs(f_new-f_old)<epsilon | norm(grad,"2")<epsilon)
      # if(norm(grad,"2")<epsilon)
    {break}
    norm_grad=norm(grad,"2")
    f_old<-f_new
    x=x+alpha*s_dir
  }
  grad<-grad_vec(x)
  norm_grad=norm(grad,"2")
  xm=x
  gradm=grad_vec(xm)
  norm_gradm=norm(gradm,"2")
  results=list(xm,norm_gradm,n_iter)
  return(results)
  
}


# now compare the standard gradient descent and the golden mean 
# gradient descent in terms of numbers of iterations needed

x_init=cbind(1,1,1)    # set the initial value for the iteration to start
epsilon=0.00001
alpha=0.1
max_iter=100

results_ap=steepest_descent_standard(x,epsilon,alpha,max_iter)
results_ap

results_ap_gs=steepest_descent_golden_search(x,epsilon,max_iter)
results_ap_gs

# increase accuracy

x_init=cbind(1,1,1)    # set the initial value for the iteration to start
epsilon=0.000000000001
alpha=0.1
max_iter=100

results_ap=steepest_descent_standard(x,epsilon,alpha,max_iter)
results_ap

results_ap_gs=steepest_descent_golden_search(x,epsilon,max_iter)
results_ap_gs


# we now modify the two functions above so as to keep track of
# of the progress of the value of the objective function
# while the gradient descent runs 


steepest_descent_standard_track<-function(x,epsilon,alpha,max_iter)
{
  f_old=func_var_multi(x)  
  f_values<-c(f_old)    # create a vector containing values of f
  n_iter=0   # counter of iteration number
  for(k in 0:max_iter)
  { 
    n_iter=n_iter+1
    grad<-grad_vec(x)
    s_dir<- -grad   # compute the gradient and set search direction -Df
    
    # compute the step length using the golden search rule
    
    
    
    x=x+alpha*s_dir
    f_new=func_var_multi(x)
    f_values<-c(f_values,f_new)  # add the new value of f to the 
                                 # vector of f values
    # if f_old and f_new are close  or
    # the gradient is close to 0 we terminate the algorithm
    
    if(abs(f_new-f_old)<epsilon | norm(grad,"2")<epsilon)
      # if(norm(grad,"2")<epsilon)
    {break}
    norm_grad=norm(grad,"2")
    f_old<-f_new
    x=x+alpha*s_dir
  }
  grad<-grad_vec(x)
  norm_grad=norm(grad,"2")
  xm=x
  gradm=grad_vec(xm)
  norm_gradm=norm(gradm,"2")
  results=list(xm,norm_gradm,n_iter,f_values)
  return(results)
}



steepest_descent_golden_search_track<-function(x,epsilon,max_iter)
{
  f_old=func_var_multi(x)
  f_values<-c(f_old)    # create a vector containing values of f
  n_iter=0
  for(k in 0:max_iter)
  {
    n_iter=n_iter+1
    grad<-grad_vec(x)
    s_dir<- -grad   # compute the gradient and set search direction -Df
    
    # compute the step length using the golden search rule
    
    golden=golden_line_search(0,100,x,s_dir,epsilon,max_iter)
    alpha<-golden[1]
    f_new<-golden[2]
    f_values<-c(f_values,f_new)  # add the new value of f to the 
    # vector of f values
    
    # if f_old and f_new are close  or
    # the gradient is close to 0 we terminate the algorithm
    
    if(abs(f_new-f_old)<epsilon | norm(grad,"2")<epsilon)
      # if(norm(grad,"2")<epsilon)
    {break}
    norm_grad=norm(grad,"2")
    f_old<-f_new
    x=x+alpha*s_dir
  }
  grad<-grad_vec(x)
  norm_grad=norm(grad,"2")
  xm=x
  gradm=grad_vec(xm)
  norm_gradm=norm(gradm,"2")
  results=list(xm,norm_gradm,n_iter,f_values)
  return(results)
  
}


# run the steepest descent and plot the progress of the 
# objective function

x_init=cbind(1,1,1)    # set the initial value for the iteration to start
epsilon=0.000000000001
alpha=0.1
max_iter=100

results_ap=steepest_descent_standard_track(x,epsilon,alpha,max_iter)
results_ap

f_values=results_ap[4]   # this will return a list




x_init=cbind(1,1,1)    # set the initial value for the iteration to start
epsilon=0.000000000001
alpha=0.1
max_iter=100

results_ap=steepest_descent_standard_track(x,epsilon,alpha,max_iter)
results_ap

f_values=results_ap[4]   # this will return a list
f_values=unlist(f_values)
n_iterations=seq(0,as.numeric(results_ap[3]))
plot(n_iterations,f_values)



x_init=cbind(1,1,1)    # set the initial value for the iteration to start
epsilon=0.000000000001
alpha=0.1
max_iter=100

results_gs=steepest_descent_golden_search_track(x,epsilon,max_iter)
results_gs

f_values=results_gs[4]   # this will return a list
f_values=unlist(f_values)
n_iterations=seq(0,as.numeric(results_gs[3]))
plot(n_iterations,f_values)